@@ -302,11 +302,75 @@
               :to-equal
               "su1 su2 sp2"))))
 
+(describe "The `buttercup-elapsed-time' function"
+  (let ((spytime (current-time)))
+    (before-each
+      (spy-on 'current-time
+              :and-call-fake
+              (lambda ()
+                (setq spytime (time-add spytime (seconds-to-time 1.5))))))
+    (it "should report elapsed time for suites"
+      (let ((suite (make-buttercup-suite)))
+        (buttercup--set-start-time suite)
+        (buttercup--set-end-time suite)
+        (expect (buttercup-elapsed-time suite)
+                :to-equal (seconds-to-time 1.5))))
+    (it "should report elapsed time for specs"
+      (let ((spec (make-buttercup-spec)))
+        (buttercup--set-start-time spec)
+        (buttercup--set-end-time spec)
+        (expect (buttercup-elapsed-time spec)
+                :to-equal (seconds-to-time 1.5))))))
+
+(defmacro with-local-buttercup (&rest body)
+  "Execute BODY with local buttercup state variables."
+  (declare (debug t) (indent defun))
+  `(let (buttercup--after-all
+         buttercup--after-each
+         buttercup--before-all
+         buttercup--before-each
+         buttercup--cleanup-functions
+         buttercup--current-suite
+         (buttercup-reporter #'ignore)
+         buttercup-suites
+         (buttercup-warning-buffer-name " *ignored buttercup warnings*"))
+     ,@body))
+
+(describe "The `buttercup--run-suite' function"
+  (before-each
+    (spy-on 'buttercup--set-start-time :and-call-through)
+    (spy-on 'buttercup--set-end-time :and-call-through))
+  (it "should set start and end time of the suite"
+    (with-local-buttercup
+      (let ((suite (make-buttercup-suite)))
+        (buttercup--run-suite suite)
+        (expect 'buttercup--set-start-time :to-have-been-called-times 1)
+        (expect (buttercup-suite-or-spec-time-started suite)
+                :not :to-be nil)
+        (expect 'buttercup--set-end-time :to-have-been-called-times 1)
+        (expect (buttercup-suite-or-spec-time-ended suite)
+                :not :to-be nil)))))
+
+(describe "The `buttercup--run-spec' function"
+    (before-each
+      (spy-on 'buttercup--set-start-time :and-call-through)
+      (spy-on 'buttercup--set-end-time :and-call-through))
+    (it "should set start and end time of the spec"
+       (with-local-buttercup
+        (let ((spec (make-buttercup-spec)))
+          (buttercup--run-spec spec)
+          (expect 'buttercup--set-start-time :to-have-been-called-times 1)
+          (expect (buttercup-suite-or-spec-time-started spec)
+                  :not :to-be nil)
+          (expect 'buttercup--set-end-time :to-have-been-called-times 1)
+          (expect (buttercup-suite-or-spec-time-ended spec)
+                  :not :to-be nil)))))
+
 ;;;;;;;;;;;;;;;;;;;;
 ;;; Suites: describe
 
 (describe "The `describe' macro"
-  (it "should expand to a simple call to the describe function"
+  (it "should expand to a simple call to the buttercup-describe function"
     (expect (macroexpand '(describe "description" (+ 1 1)))
             :to-equal
             '(buttercup-describe "description" (lambda () (+ 1 1)))))
@@ -461,19 +525,36 @@
   (it "expands directly to a function call"
     (expect (macroexpand '(xdescribe "bla bla" (+ 1 1)))
             :to-equal
-            '(buttercup-xdescribe "bla bla" (lambda () (+ 1 1))))))
+            '(buttercup-describe "bla bla"
+                                 (lambda ()
+                                   (signal 'buttercup-pending "PENDING")))))
 
-(describe "The `buttercup-xdescribe' function"
-  (it "should be a no-op"
-    (expect (buttercup-xdescribe
-             "bla bla"
-             (lambda () (error "Should not happen")))
-            :not :to-throw))
+  (it "changes contained it-specs to pending specs"
+    (expect (macroexpand-all
+             '(xdescribe "bla bla"
+                (let ((a 1) b (c 2) (d (it "nested" (+ 1 1))))
+                  (it "spec1" (+ 1 1))
+                  (describe "inner suite"
+                    (it "inner spec"))
+                  (xit "spec2" (+ 1 1)))))
+            :to-equal
+            '(buttercup-describe
+              "bla bla"
+              #'(lambda ()
+                  (buttercup-xit "nested")
+                  (buttercup-xit "spec1")
+                  (buttercup-describe
+                   "inner suite"
+                   #'(lambda ()
+                       (buttercup-xit "inner spec")
+                       (signal 'buttercup-pending "PENDING")))
+                  (buttercup-xit "spec2")
+                  (signal 'buttercup-pending "PENDING")))))
 
   (it "should add a pending suite"
     (let ((buttercup--current-suite nil)
           (buttercup-suites nil))
-      (buttercup-xdescribe
+      (xdescribe
        "bla bla"
        (lambda () nil))
       (expect (buttercup-suite-status (car buttercup-suites))
@@ -513,7 +594,15 @@
       (expect (buttercup-spec-status
                (car (last (buttercup-suite-children
                            buttercup--current-suite))))
-              :to-be 'pending))))
+              :to-be 'pending)))
+
+  (it "should set the failure description to PENDING"
+    (let ((buttercup--current-suite (make-buttercup-suite))
+          spec)
+      (buttercup-xit "bla bla")
+      (setq spec (car (buttercup-suite-children buttercup--current-suite)))
+      (buttercup--update-with-funcall spec (buttercup-spec-function spec))
+      (expect (buttercup-suite-or-spec-failure-description spec) :to-equal "PENDING"))))
 
 ;;;;;;;;;
 ;;; Spies
@@ -823,10 +912,16 @@
                 "spec")))
 
     (describe "on the spec-done event"
-      (it "should simply emit a newline for a passed spec"
+      (it "should print no status tag for a passed spec"
+        (buttercup--set-start-time spec)
+        (setf (buttercup-spec-failure-description spec) "DONTSHOW")
+        (buttercup--set-end-time spec)
         (buttercup-reporter-batch 'spec-done spec)
 
-        (expect 'buttercup--print :to-have-been-called-with "\n"))
+        (expect (mapconcat (apply-partially #'apply #'format)
+                           (spy-calls-all-args 'buttercup--print)
+                           "")
+                :to-match "^\\s-*([0-9]+\\.[0-9]+\\(h\\|m\\|m?s\\))\n$"))
 
       (it "should say FAILED for a failed spec"
         (setf (buttercup-spec-status spec) 'failed)
@@ -834,13 +929,38 @@
         (let ((buttercup-reporter-batch--failures nil))
           (buttercup-reporter-batch 'spec-done spec))
 
-        (expect 'buttercup--print :to-have-been-called-with "  FAILED\n"))
+        (expect (mapconcat (apply-partially #'apply #'format)
+                           (spy-calls-all-args 'buttercup--print)
+                           "")
+                :to-match "FAILED\\(\\s-+.*\\)?\n$"))
+
+      (it "should output the failure-description for a pending spec"
+        (setf (buttercup-spec-status spec) 'pending
+              (buttercup-spec-failure-description spec) "DESCRIPTION")
+        (let ((buttercup-reporter-batch--failures nil))
+          (buttercup-reporter-batch 'spec-done spec))
+        (expect (mapconcat (apply-partially #'apply #'format)
+                           (spy-calls-all-args 'buttercup--print) "")
+                :to-match "DESCRIPTION\\(\\s-+.*\\)?\n$"))
 
       (it "should throw an error for an unknown spec status"
         (setf (buttercup-spec-status spec) 'unknown)
 
         (expect (buttercup-reporter-batch 'spec-done spec)
-                :to-throw)))
+                :to-throw))
+
+      (it "should print the elapsed time for all specs"
+        (dolist (state '(pending failed passed))
+          (setq spec (make-buttercup-spec :description "spec" :status state :parent child-suite))
+          (buttercup--set-start-time spec)
+          (buttercup--set-end-time spec)
+          (let ((buttercup-reporter-batch--failures nil))
+            (buttercup-reporter-batch 'spec-done spec))
+
+          (expect (mapconcat (apply-partially #'apply #'format)
+                             (spy-calls-all-args 'buttercup--print)
+                             "")
+                  :to-match "([0-9]+\\.[0-9]+\\(h\\|m\\|m?s\\))"))))
 
     (describe "on the suite-done event"
       (it "should emit a newline at the end of the top-level suite"
@@ -859,17 +979,49 @@
       (it "should handle the end event"
         (buttercup-reporter-batch 'buttercup-done nil))
 
-      (it "should raise an error if at least one spec failed"
+      (it "should not raise any error even if a spec failed"
         (setf (buttercup-spec-status spec) 'failed)
 
         (expect (buttercup-reporter-batch 'buttercup-done (list spec))
-                :to-throw)))
+                :not :to-throw)))
 
     (describe "on an unknown event"
       (it "should raise an error"
         (expect (buttercup-reporter-batch 'unknown-event nil)
                 :to-throw)))))
 
+(describe "The `buttercup-run' function"
+  :var (buttercup-suites parent-suite child-suite spec)
+  (cl-flet (((reporter (event arg) (ignore event arg))))
+    (before-each
+      (setq parent-suite (make-buttercup-suite :description "parent-suite")
+            child-suite (make-buttercup-suite :description "child-suite")
+            spec (make-buttercup-spec :description "spec")
+            buttercup-suites (list parent-suite))
+      (buttercup-suite-add-child parent-suite child-suite)
+      (buttercup-suite-add-child child-suite spec)
+      (spy-on 'reporter))
+    (it "should raise an error if at least one spec failed"
+      (setf (buttercup-spec-status spec) 'failed)
+      (cl-letf (((symbol-function 'buttercup--run-suite) #'ignore)
+                (buttercup-reporter 'reporter))
+        (expect (buttercup-run) :to-throw)))
+    (it "should call the reporter twice with events buttercup-started and -done"
+      (cl-letf (((symbol-function 'buttercup--run-suite) #'ignore)
+                (buttercup-reporter 'reporter))
+        (expect (buttercup-run) :not :to-throw)
+        (expect 'reporter :to-have-been-called-times 2)
+        (expect 'reporter :to-have-been-called-with 'buttercup-started buttercup-suites)
+        (expect 'reporter :to-have-been-called-with 'buttercup-done buttercup-suites)))
+    (it "should call `buttercup--run-suite once per suite"
+      (cl-flet (((runner (suite) (ignore suite))))
+        (spy-on 'runner)
+        (cl-letf (((symbol-function 'buttercup--run-suite) #'runner)
+                  (buttercup-reporter 'reporter)
+                  (buttercup-suites (make-list 5 parent-suite)))
+          (expect (buttercup-run) :not :to-throw)
+          (expect 'runner :to-have-been-called-times 5))))))
+
 (describe "The `buttercup--print' function"
   (before-each
     (spy-on 'send-string-to-terminal))
@@ -960,6 +1112,9 @@
         (expect (length (cdr specs)) :to-equal 1)
         (expect (cl-caadr specs) :to-equal "should fontify special keywords")))))
 
+;; Local Variables:
+;; indent-tabs-mode: nil
+;; sentence-end-double-space: nil
+;; End:
 (provide 'test-buttercup)
-
 ;;; test-buttercup.el ends here
